# coding: utf-8
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

=begin
Copyright (c) 2010 KISHIMOTO, Makoto

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

Version

  0.0.1  --  2010/Dec/8  first version

History

  0.0.1  --  2010/Dec/8  first version

=end

require "sqlite3"

module JTokenizer

class TrieDB
	def self.open
		dir = File.dirname __FILE__
		SQLite3::Database.new "#{dir}/trie.db"
	end
end

class TrieDBNode
	def self.root_node db
		new db, "root"
	end

	def initialize db, node_id
		@db = db
		@node_id = node_id
		@tango = nil
		@skipstr = nil
		@children = nil
	end

	def tango
		if @tango then
			return @tango[0]
		end
		sql = "select tango from tbl_attrs where id == :id"
		r = @db.execute sql, :id => @node_id
		(@tango = if r[0][0] == 0 then [false] else [true] end)[0]
	end

	def skipstr
		if @skipstr then
			return @skipstr
		end
		sql = "select skipstr from tbl_attrs where id == :id"
		r = @db.execute sql, :id => @node_id
		@skipstr = r[0][0]
	end

	def children
		if @children then
			return @children
		end
		@children = Chains.new @db, @node_id
	end
end

class Chains
	def initialize db, node_id
		@db = db
		@node_id = node_id
		@chains = nil
	end

	def key? s
		if @chains then
			@chains.key? s
		else
			sql = "select key, chain from tbl_chains where id == :id"
			r = @db.execute sql, :id => @node_id
			@chains = Hash[*r.flatten]
			@chains.key? s
		end
	end

	def [] s
		r = if @chains then
			@chains[s]
		else
			key? s and @chains[s]
		end
		r and case r
		when TrieDBNode then
			r
		when String then
			@chains[s] = TrieDBNode.new @db, r
		else
			raise
		end
	end
end

STR_SRC =
	"　！”＃＄％＆’（）＊＋，‐—\u2212．／０１２３４５６７８９：；＜＝＞？" +
	"＠ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ［＼￥］＾＿" +
	"｀ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ｛｜｝〜￣"

STR_ALT =
	%q[ !"#$%&'()*+,\-\-\-./0123456789:;<=>?] +
	'@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\\\\\\\]^_' +
	'`abcdefghijklmnopqrstuvwxyz{|}~~'

def self.zenkaku2hankaku s
	s.tr STR_SRC, STR_ALT
end

def self.build_db db, trie
	root_node = trie
	node_fifo = []
	root_node.children.each_pair {|key, node|
		node_fifo.push node
	}
	db.execute <<SQL_END
create table tbl_attrs (
  id text,
  tango integer,
  skipstr text,
  primary key(id)
);
SQL_END
	db.execute <<SQL_END
create table tbl_chains (
  id text,
  key text,
  chain text,
  primary key(id, key)
);
SQL_END
	#count = 0
	db.transaction
	begin
		loop {
			#if count % 100 == 0 then
			#	p count
			#end
			unless node = node_fifo.shift then
				break
			end
			node_id = node.object_id.to_s
			sql = "insert into tbl_attrs values (:id, :tango, :skipstr)"
			db.execute sql, :id => node_id, :tango => if node.tango then 1 else 0 end, :skipstr => node.skipstr

			node.children.each_pair {|key, node|
				sql = "insert into tbl_chains values (:id, :key, :chain)"
				db.execute sql, :id => node_id, :key => key, :chain => node.object_id.to_s

				node_fifo.push node
			}
			#count += 1
		}
	rescue
		db.rollback
	else
		db.commit
	end

	# make root entry
	node = root_node
	node_id = "root"
	db.transaction
	begin
		sql = "insert into tbl_attrs values (:id, :tango, :skipstr)"
		db.execute sql, :id => node_id, :tango => if node.tango then 1 else 0 end, :skipstr => node.skipstr

		node.children.each_pair {|key, node|
			sql = "insert into tbl_chains values (:id, :key, :chain)"
			db.execute sql, :id => node_id, :key => key, :chain => node.object_id.to_s
		}
	rescue
		db.rollback
	else
		db.commit
	end
end

class JTokenizer
	def initialize
		@db = TrieDB.open
	end

	def close
		@db.close
	end

	def root_node
		@root_node ||= TrieDBNode.root_node @db
	end

	def prefixes s
		self.class.match_sub [], "", s, root_node
	end

	def self.match_sub result, buf, rest, node
		# result 途中の結果を入れる Array 。複数個のマッチの途中の場合途中経過が入っている
		# buf トークンの一部として既に認識した部分
		# rest 残りの文字列
		# node trie のカレントノード
		c = nil
		s = ""
		if (sk = node.skipstr).empty? then
			c = rest[0, 1]
		else
			# skipstr が空でない場合は、マッチ候補パターンが確定している
			sk_ = rest[0, sk.length]
			if sk == sk_ then
				# スキップパターンがマッチ
				s = sk
				c = rest[sk.length, 1]
			else
				# skipstr が空でなくて、マッチしなかった場合は trie の袋小路なので
				# result を返して終わり
				return result
			end
		end

		if node.tango then
			# このノードは単語の終わりなので、buf + s の内容はトークンである
			result << (buf + s)

			# もっと長い単語の一部の可能性もあるのでここでは return しない
		end

		if node.children.key? c then
			# ノードに、先頭文字にマッチしている子がある場合はそれでマッチ（再帰）
			s << c
			buf << s
			return match_sub result, buf, rest[s.length .. -1], node.children[c]
		else
			# ここに来た場合も trie の袋小路
			return result
		end

		# ここには来ない
		raise
	end
end

end
